#ifndef _zxidvers_h
#define _zxidvers_h
#define ZXID_VERSION 0x000115
#define ZXID_REL "1.15"
#define ZXID_COMPILE_DATE "1382893016"
#define ZXID_REV "$Id: 1.14-4-gb587add 20131027-163722 sampo@ $"
#endif
